# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython


cdef class DebugOptions(object):
    cdef public bint DUMP_PRE_ERROR_STATE
    cdef public bint DUMP_EXCEPTIONS
    cdef public bint DUMP_SCHEDULE_TASK
    cdef public bint DUMP_CONTINUE_TASK
    cdef public bint DUMP_SCHEDULE_BATCH
    cdef public bint DUMP_FLUSH_BATCH
    cdef public bint DUMP_DEPENDENCIES
    cdef public bint DUMP_COMPUTED
    cdef public bint DUMP_NEW_TASKS
    cdef public bint DUMP_YIELD_RESULTS
    cdef public bint DUMP_QUEUED_RESULTS
    cdef public bint DUMP_CONTEXTS
    cdef public bint DUMP_SYNC
    cdef public bint DUMP_STACK
    cdef public bint DUMP_SCHEDULER_STATE
    cdef public bint DUMP_SYNC_CALLS
    cdef public bint COLLECT_PERF_STATS

    cdef public float SCHEDULER_STATE_DUMP_INTERVAL
    cdef public int DEBUG_STR_REPR_MAX_LENGTH
    cdef public object STACK_DUMP_LIMIT
    cdef public int MAX_TASK_STACK_SIZE

    cdef public object ENABLE_COMPLEX_ASSERTIONS

    cpdef DUMP_ALL(self, value=?)

cdef public DebugOptions options



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time


class Profiler(object):
    def __init__(self, label):
        self.label = label
        self.diff = 0

    def __enter__(self):
        print('Entering %s:' % self.label)
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        self.diff = time.time() - self.start
        print('{0} took {1:.3f}s'.format(self.label, self.diff))



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython

cimport async_task


@cython.locals(active_task=async_task.AsyncTask)
cdef async_task.AsyncTask enter_context(object context)
cdef leave_context(object context, async_task.AsyncTask active_task)


cdef class NonAsyncContext(object):
    cpdef NonAsyncContext __enter__(self)
    cpdef __exit__(self, ty, val, tb)
    cpdef pause(self)
    cpdef resume(self)

cdef class AsyncContext(object):
    cdef public async_task.AsyncTask _active_task
    cpdef AsyncContext __enter__(self)
    cpdef __exit__(self, ty, val, tb)



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__doc__ = """

Tests that when using multiple inheritance, both parent classes' methods can be called.

See the implementation of DecoratorBase.__get__ for how this works.

"""

from asynq import asynq, result


called = {}


class Parent1(object):
    @asynq()
    def method(self):
        called['Parent1'] = True


class Parent2(object):
    @asynq()
    def method(self):
        called['Parent2'] = True


class Child(Parent1, Parent2):
    @asynq()
    def method(self):
        yield super(Child, self).method.asynq()
        yield Parent2.method.asynq(self)
        called['Child'] = True


def test():
    @asynq()
    def inner():
        instance = Child()
        yield instance.method.asynq()
        result(None); return

    inner()
    assert called['Parent1']
    assert called['Parent2']
    assert called['Child']



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from qcore.debug import get_bool_by_mask, set_by_mask


class DebugOptions(object):
    """
    All async debug options are stored in this structure.

    The class itself is necessary mainly by performance reasons:
    we want to avoid dictionary lookups on checks of debug options.

    """
    def __init__(self):
        self.DUMP_PRE_ERROR_STATE  = False
        self.DUMP_EXCEPTIONS       = False
        self.DUMP_SCHEDULE_TASK    = False
        self.DUMP_CONTINUE_TASK    = False
        self.DUMP_SCHEDULE_BATCH   = False
        self.DUMP_FLUSH_BATCH      = False
        self.DUMP_DEPENDENCIES     = False
        self.DUMP_COMPUTED         = False
        self.DUMP_NEW_TASKS        = False
        self.DUMP_YIELD_RESULTS    = False
        self.DUMP_QUEUED_RESULTS   = False
        self.DUMP_CONTEXTS         = False
        self.DUMP_SYNC             = False
        self.DUMP_STACK            = False  # When it's meaningful, e.g. on batch flush
        self.DUMP_SCHEDULER_STATE  = False
        self.DUMP_SYNC_CALLS       = False
        self.COLLECT_PERF_STATS    = False

        self.SCHEDULER_STATE_DUMP_INTERVAL = 1     # In seconds
        self.DEBUG_STR_REPR_MAX_LENGTH     = 240   # In characters, 0 means infinity
        self.STACK_DUMP_LIMIT              = 0     # In frames, None means infinity

        self.ENABLE_COMPLEX_ASSERTIONS = True

    def DUMP_ALL(self, value=None):
        if value is None:
            return get_bool_by_mask(self, "DUMP_")
        set_by_mask(self, "DUMP_", value)


options = DebugOptions()
globals()['options'] = options



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
from cpython.ref cimport PyObject

cimport qcore.events as core_events
cimport qcore.helpers as core_helpers
cimport qcore.inspection as core_inspection
cimport futures
cimport scheduler
cimport _debug
cimport profiler


cdef _debug.DebugOptions _debug_options
cdef object _futures_none
cdef object _none_future


cdef class AsyncTask(futures.FutureBase):
    cdef public long iteration_index
    cdef public object fn
    cdef public object args
    cdef public object kwargs
    cdef public str _name
    cdef public AsyncTask creator
    cdef public object _generator
    cdef public object _last_value
    cdef public object _frame
    cdef public object perf_stats
    cdef public list _dependencies
    cdef public list _contexts
    cdef public bint _contexts_active
    cdef public bint _dependencies_scheduled
    cdef public int _total_time
    cdef public int _id

    cdef bint is_blocked(self) except -1
    cdef bint can_continue(self) except -1

    cpdef _compute(self)
    cpdef _computed(self)
    cpdef dump_perf_stats(self)
    cpdef collect_perf_stats(self)
    cpdef to_str(self)

    cdef _continue(self)
    cdef inline object _continue_on_generator(self, value, error)

    @cython.locals(scheduler=scheduler.TaskScheduler)
    cdef inline object _accept_yield_result(self, result)
    cdef inline object _accept_error(self, error)

    cdef _queue_exit(self, result)
    cdef _queue_throw_error(self, error)

    cpdef list traceback(self)

    cpdef dump(self, int indent=?)

    # Contexts support

    cdef _enter_context(self, context)
    cdef _leave_context(self, context)
    @cython.locals(i=int)
    cdef _pause_contexts(self)
    @cython.locals(i=int)
    cdef _resume_contexts(self)


@cython.locals(length=int, future=futures.FutureBase, tpl=tuple, lst=list, dct=dict)
cdef object unwrap(object value)

cpdef extract_futures(object value, list result)

cdef tuple _empty_tuple
cdef dict _empty_dictionary



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import asynq, Future, scheduler

values = {}


@asynq(pure=True)
def get(key):
    global values
    value = values.get(key)
    print('Get %s -> %s' % (str(key), str(value)))
    return value


@asynq(pure=True)
def set(key, value):
    global values
    values[key] = value
    print('Set %s <- %s' % (str(key), str(value)))


@asynq(pure=True)
def get_and_set(key_from, key_to, depends_on):
    yield depends_on
    value = yield get(key_from)
    yield set(key_to, value)


@asynq(pure=True)
def order_test():
    global values
    values = {}
    prev_task = set(0, 'value')
    tasks = []
    for i in range(0, 10):
        task = get_and_set(i, i + 1, prev_task)  # No need to yield!
        prev_task = task
        tasks.append(task)

    assert len(values) == 0  # Nothing is executed yet!

    yield tasks
    assert len(values) == 11  # Done at this point

    yield  # Nothing happens here
    assert len(values) == 11  # Done at this point

def test():
    order_test()()



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import async, scheduler
from qcore.asserts import assert_is


@async()
def outer_async():
    """Test that we get the correct active task from the scheduler.

    Even when the execution of one task gets interrupted by a synchonous call to another async
    function, the scheduler retains the correct active task.

    """
    active_task = scheduler.get_active_task()
    inner_async()
    assert_is(active_task, scheduler.get_active_task())


@async()
def inner_async():
    pass


def test_active_task():
    outer_async()



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import asynq, result
from .helpers import Profiler

counter = 0


def test():
    global counter

    @asynq(pure=True)
    def incr():
        global counter
        counter += 1
        print("Counter: %i" % counter)
        result(counter); return
        yield

    def sync_incr():
        global counter
        counter += 1
        print("Counter: %i" % counter)
        return counter

    @asynq(pure=True)
    def test_async():
        global counter
        try:
            print('In try block.')
            yield incr()
            result((yield incr()))  # ; return
            assert False, "Must not reach this point!"
        except BaseException as e:
            print('In except block, e = ' + repr(e))
            assert sync_incr() == 3
            if isinstance(e, GeneratorExit):
                raise
            assert False, "Must not reach this point!"
        finally:
            print('In finally block.')
            assert sync_incr() == 4

    with Profiler('test_stop()'):
        r = test_async()()
        assert r == 2
        assert counter == 4
    print()



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import six
import sys

import qcore.helpers as core_helpers
import qcore.inspection as core_inspection
import qcore.errors as core_errors

from . import debug
from . import futures
from . import profiler
from . import _debug

import asynq

__traceback_hide__ = True

_debug_options = _debug.options
_futures_none = futures._none
_none_future = futures.none_future

MAX_DUMP_INDENT = 40


class AsyncTaskCancelledError(GeneratorExit):
    pass


class AsyncTaskResult(GeneratorExit):
    def __init__(self, result):
        GeneratorExit.__init__(self, "AsyncTaskResult")
        self.result = result


class AsyncTask(futures.FutureBase):
    """Asynchronous task implementation.
    Uses passed generator (normally produced by generator method)
    to implement asynchronous "await" behavior via 'yield' keyword.

    """

    def __init__(self, generator, fn, args, kwargs):
        super(AsyncTask, self).__init__()
        if _debug_options.ENABLE_COMPLEX_ASSERTIONS:
            assert core_inspection.is_cython_or_generator(generator), \
                'generator is expected, but %s is provided' % generator
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.iteration_index = 0
        self._generator = generator
        self._frame = None
        self._last_value = None
        self._dependencies = []
        self._contexts = []
        self._contexts_active = False
        self._dependencies_scheduled = False
        self._total_time = 0
        self._name = None
        self.perf_stats = {}
        self.creator = asynq.scheduler.get_active_task()
        if _debug_options.DUMP_NEW_TASKS:
            debug.write('@async: new task: %s, created by %s' %
                (debug.str(self), debug.str(self.creator)))
        if _debug_options.COLLECT_PERF_STATS:
            self._id = profiler.incr_counter()

    def can_continue(self):
        """Indicates whether this async task has more steps to execute.

        Task can't continue, if its underlying generator is disposed.

        """
        return self._generator is not None

    def is_blocked(self):
        """Indicates whether this async task is currently blocked.
        Blocked tasks are tasks waiting for completion of other
        tasks or batches.

        """
        for dependency in self._dependencies:
            if not dependency.is_computed():
                return True
        return False

    def _compute(self):
        if _debug_options.DUMP_SYNC_CALLS:
            if self.creator is not None:
                debug.write('@async: %s called synchronously from %s'
                    % (debug.str(self), debug.str(self.creator)))
        # Forwards the call to task scheduler
        asynq.scheduler.get_scheduler().wait_for(self)
        # No need to assign a value/error here, since
        # _continue method (called by TaskScheduler) does this.

    def dump_perf_stats(self):
        self.perf_stats['time_taken'] = self._total_time
        profiler.append(self.perf_stats)

    def to_str(self):
        if self._name is None:
            try:
                self._name = '%06d.%s(%r %r)' % (
                    self._id,
                    core_inspection.get_full_name(self.fn),
                    self.args,
                    self.kwargs
                )
            except RuntimeError:
                self._name = '%06d.%s' % (self._id, core_inspection.get_full_name(self.fn))
        return self._name

    def collect_perf_stats(self):
        self.perf_stats = {
            'name': self.to_str(),
            'num_deps': len(self._dependencies),
            'dependencies': [
                (t.to_str(), t._total_time)
                for t in self._dependencies if isinstance(t, AsyncTask)
            ]
        }

    def _computed(self):
        try:
            if self._generator is not None:
                self._generator.close()
                self._generator = None
            if _debug_options.COLLECT_PERF_STATS is True:
                self.collect_perf_stats()
        finally:
            # super() doesn't work in Cython-ed version here
            self._dependencies = []
            self._last_value = None
            futures.FutureBase._computed(self)
            error = self.error()

    def _continue(self):
        # Obvious optimization: we don't need to return from
        # this method, if we still can go further after yield.
        # So we return only if there are dependencies or
        # current task is computed.
        while True:
            try:
                value = unwrap(self._last_value)
                error = None
            except BaseException as e:
                value = None
                error = e
            try:
                self._accept_yield_result(self._continue_on_generator(value, error))
            except StopIteration as error:  # Most frequent, so it's the first one
                try:
                    # We're on a Python version that supports adding a value to StopIteration
                    return_value = error.value
                except AttributeError:
                    # This means there was no asynq.result() call, so the value of
                    # this task should be None
                    return_value = None
                self._queue_exit(return_value)
            except GeneratorExit as error:
                error_type = type(error)
                if error_type is AsyncTaskResult:
                    self._queue_exit(error.result)
                elif error_type is AsyncTaskCancelledError:
                    self._accept_error(error)
                else:
                    self._queue_exit(None)
            except BaseException as error:
                self._accept_error(error)

            if self.is_computed():
                return
            if len(self._dependencies) > 0:
                return

    def _continue_on_generator(self, value, error):
        try:
            if error is None:
                if self._generator is None:
                    raise StopIteration()
            elif self._generator is None:
                raise error

            self.iteration_index += 1
            if error is None:
                return self._generator.send(value)
            else:
                self._last_value = None  # No need to keep it further
                self._frame = debug.get_frame(self._generator)
                if hasattr(error, '_task'):
                    return self._generator.throw(error._type_, error, error._traceback)
                else:
                    return self._generator.throw(type(error), error)
        except (StopIteration, GeneratorExit):
            # Returning leads to a StopIteration exception, which is
            # handled here. In this case we shouldn't need to extract frame
            # info, because that is slow.
            self._generator = None
            raise
        except:
            # Any exception means generator is closed at this point,
            # since it fall through all exception handlers there;
            # "return" leads to StopIteration exception, so it's
            # handled here as well.
            # If the task failed, we want to save the frame info here so that the traceback can
            # show where in the async task the failure happened. However, if the error was thrown
            # into the generator, we'll already have set the frame info.
            if self._frame is None:
                tb = sys.exc_info()[2]

                while tb.tb_next is not None:
                    tb = tb.tb_next
                self._frame = tb.tb_frame
            self._generator = None
            raise

    def _accept_yield_result(self, result):
        if _debug_options.DUMP_YIELD_RESULTS:
            debug.write('@async: yield: %s -> %s' % (debug.str(self), debug.repr(result)))
        self._last_value = result
        extract_futures(result, self._dependencies)

    def _accept_error(self, error):
        if self._value is not _futures_none:  # is_computed(), but faster
            # We can't change the value in this case, so we can just return here.
            pass
        else:
            # when there is no _task it means that this is the bottommost level of the async
            # task. We must attach the traceback as soon as possible
            if not hasattr(error, '_task'):
                error._task = self
                core_errors.prepare_for_reraise(error)
            else:
                # when we already have the _task on the error, it means that
                # some child generator of ours had an error.
                # now, we are storing the _traceback on the error. we use this so we can
                # raise it with that exact traceback later.
                # now, we when do raise it, the upper level gets a new traceback
                # with the curent level's traceback connected via a linked list pointer.
                # known as tb_next in traceback object.
                # this is really important. if we keep updating this traceback,
                # we can glue all the different tasks' tracebacks and make it look like
                # the error came from there.
                error._traceback = sys.exc_info()[2]

            if _debug_options.DUMP_EXCEPTIONS:
                debug.dump_error(error)
            # Must queue, since there can be captured dependencies to resolve
            self._queue_throw_error(error)

    def _queue_exit(self, result):
        # Result must be unwrapped first.
        # All of the dependencies must have been computed at this point.
        if self._value is not _futures_none:  # is_computed(), but faster
            raise futures.FutureIsAlreadyComputed(self)
        if self._generator is not None:
            self._generator.close()
            self._generator = None
        if _debug_options.DUMP_QUEUED_RESULTS:
            debug.write('@async: queuing exit: %s <- %s' % (debug.repr(result), debug.str(self)))
        self.set_value(result)

    def _queue_throw_error(self, error):
        if self._value is not _futures_none:  # is_computed(), but faster
            raise futures.FutureIsAlreadyComputed(self)
        if _debug_options.DUMP_QUEUED_RESULTS:
            debug.write('@async: queuing throw error: %s <-x- %s' % (debug.repr(error), debug.str(self)))
        self.set_error(error)

    def traceback(self):
        try:
            self_str = self._traceback_line()
        except Exception:
            # If _traceback_line failed for whatever reason (e.g. there is no correct frame),
            # fall back to __str__ so that we can still provide useful information for debugging
            self_str = core_helpers.safe_str(self)
        if self.creator is None:
            return [self_str]
        result = self.creator.traceback()
        result.append(self_str)
        return result

    def _traceback_line(self):
        frame = self._frame
        if frame is None and self._generator is not None:
            frame = debug.get_frame(self._generator)

        if frame is not None:
            frame_info = inspect.getframeinfo(frame)
            template = '''File "%(file)s", line %(lineno)s, in %(funcname)s
    %(codeline)s'''
            return template % {
                'file': frame_info.filename,
                'lineno': frame_info.lineno,
                'funcname': frame_info.function,
                'codeline': '\n'.join(frame_info.code_context).strip()
            }
        else:
            return str(self)

    def __str__(self):
        fn_str = core_inspection.get_function_call_str(self.fn, self.args, self.kwargs)
        name = '@asynq %s' % fn_str

        # we subtract one because by the time the stacktrace is printed
        # the iteration_index has already been incremented
        if self.iteration_index - 1 == 0:
            step = 'before 1st yield'
        else:
            step = 'passed yield #%i' % (self.iteration_index - 1)

        if self.is_computed():
            status = 'computed, '
            if self.error() is None:
                status += '= ' + repr(self.value())
            else:
                status += 'error = ' + repr(self.error())
        else:
            if self.is_blocked():
                status = 'blocked x%i' % len(self._dependencies)
            elif self.can_continue():
                status = 'waiting'
            else:
                status = 'almost finished (generator is closed)'
        return '%s (%s, %s)' % (name, status, step)

    def dump(self, indent=0):
        if indent > MAX_DUMP_INDENT:
            debug.write('...', indent + 1)
            return
        debug.write(debug.str(self), indent)
        if self._dependencies:
            debug.write('Dependencies:', indent + 1)
            for dependency in self._dependencies:
                dependency.dump(indent + 2)  # Recursive
                # debug.write(debug.str(dependency), indent + 2)
        else:
            debug.write('No dependencies.', indent + 1)

    # Contexts support

    def _enter_context(self, context):
        if _debug_options.DUMP_CONTEXTS:
            debug.write('@async: +context: %s' % debug.str(context))
        self._contexts.append(context)

    def _leave_context(self, context):
        if _debug_options.DUMP_CONTEXTS:
            debug.write('@async: -context: %s' % debug.str(context))
        self._contexts.remove(context)

    def _pause_contexts(self):
        if not self._contexts_active:
            return
        self._contexts_active = False
        contexts = self._contexts
        i = len(contexts) - 1
        # execute each pause() in a try/except and if 1 or more of them
        # raise an exception, then save the last exception raised so that it
        # can be re-raised later. We re-raise the last exception to make the
        # behavior consistent with __exit__.
        error = None
        while i >= 0:
            try:
                contexts[i].pause()
            except BaseException as e:
                error = e
                core_errors.prepare_for_reraise(error)
            i -= 1
        if error is not None:
            self._accept_error(error)

    def _resume_contexts(self):
        if self._contexts_active:
            return
        self._contexts_active = True
        i = 0
        contexts = self._contexts
        l = len(contexts)
        # same try/except deal as with _pause_contexts, but in this case
        # we re-raise the first exception raised.
        error = None
        while i < l:
            try:
                contexts[i].resume()
            except BaseException as e:
                if error is None:
                    error = e
                    core_errors.prepare_for_reraise(error)
            i += 1
        if error is not None:
            self._accept_error(error)


def unwrap(value):
    """
    'Unwraps' the provided arguments by converting:

    * ``FutureBase`` objects to their values;
      consequently, ``AsyncTask`` objects are also
      converted to their values
    * Tuples containing ``FutureBase`` or ``AsyncTask``
      objects are converted to tuples containing their
      values
    * Tuples are processed recursively.

    """
    if value is None:  # Special case
        return None
    elif isinstance(value, futures.FutureBase):
        future = value
        return future.value()
    elif type(value) is tuple:
        tpl = value
        # Simple perf. optimization
        length = len(tpl)
        if length <= 1:
            if length == 0:
                return ()
            return (unwrap(tpl[0]),)
        elif length == 2:
            return (unwrap(tpl[0]), unwrap(tpl[1]))
        else:
            result = []
            for item in tpl:
                result.append(unwrap(item))
            return tuple(result)
    elif type(value) is list:
        lst = value
        return [unwrap(item) for item in lst]
    elif type(value) is dict:
        dct = value
        return {key: unwrap(value) for key, value in six.iteritems(dct)}
    else:
        raise TypeError(
            "Cannot unwrap an object of type '%s': only futures and None are allowed." %
            type(value)
        )


# Private part

_empty_tuple = tuple()
_empty_dictionary = dict()
globals()['_empty_tuple'] = _empty_tuple
globals()['_empty_dictionary'] = _empty_dictionary

def extract_futures(value, result):
    """Enumerates all the futures inside a particular value."""
    if value is None:
        pass
    elif isinstance(value, futures.FutureBase):
        result.append(value)
    elif type(value) is tuple or type(value) is list:
        # backwards because tasks are added to a stack, so the last one executes first
        i = len(value) - 1
        while i >= 0:
            extract_futures(value[i], result)
            i -= 1
    elif type(value) is dict:
        for item in six.itervalues(value):
            extract_futures(item, result)
    return result



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
from cpython.ref cimport PyObject

cimport qcore.events as core_events
cimport qcore.helpers as core_helpers
cimport _debug

cdef _debug.DebugOptions _debug_options
cdef core_helpers.MarkerObject _none


@cython.auto_pickle(False)
cdef class FutureBase(object):
    cdef public object _value
    cdef public object _error
    cdef bint _in_repr
    cdef public core_events.EventHook on_computed

    cpdef object value(self)
    cpdef set_value(self, value)
    cpdef reset_unsafe(self)

    cpdef object error(self)
    cpdef set_error(self, error)

    cpdef bint is_computed(self) except -1
    cpdef _compute(self)
    cpdef _computed(self)

    cdef inline raise_if_error(self)

    cpdef dump(self, int indent=?)


@cython.auto_pickle(False)
cdef class Future(FutureBase):
    cdef public object _value_provider
    cpdef _compute(self)


@cython.auto_pickle(False)
cdef class ConstFuture(FutureBase):
    pass

@cython.auto_pickle(False)
cdef class ErrorFuture(FutureBase):
    pass



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__doc__ = """

Asynq is a framework for asynchronous programming in Python.

It supports futures, batching, and asynchronous contexts.

"""
import sys

from . import debug
from . import futures
from . import batching
from . import async_task
from . import scheduler
from . import decorators
from . import utils
from . import contexts
from . import scoped_value
from . import tools
from . import profiler
from .futures import FutureBase, Future, FutureIsAlreadyComputed, none_future, ConstFuture, \
    ErrorFuture
from .batching import BatchBase, BatchItemBase, BatchingError, BatchCancelledError
from .async_task import AsyncTask, AsyncTaskCancelledError, AsyncTaskResult
from .scheduler import TaskScheduler, get_scheduler, get_active_task, AsyncTaskError
from .decorators import asynq, async_proxy, has_async_fn, \
    is_pure_async_fn, is_async_fn, get_async_fn, get_async_or_sync_fn, async_call, \
    make_async_decorator
if sys.version_info <= (3, 7):
    exec('from .decorators import async')
from .utils import result
from .contexts import NonAsyncContext, AsyncContext
from .scoped_value import AsyncScopedValue, async_override
from . import mock_ as mock
from .generator import END_OF_GENERATOR, async_generator, list_of_generator, take_first, Value

debug.sync = batching.sync
debug.attach_exception_hook()




# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import contextlib
import sys
import qcore
import linecache
import traceback
import logging
from sys import stderr, stdout

from . import _debug

options = _debug.options  # Must be the same object

options.DUMP_PRE_ERROR_STATE  = True
options.DUMP_EXCEPTIONS       = False
options.DUMP_SCHEDULE_TASK    = False
options.DUMP_CONTINUE_TASK    = False
options.DUMP_SCHEDULE_BATCH   = False
options.DUMP_FLUSH_BATCH      = False
options.DUMP_DEPENDENCIES     = False
options.DUMP_COMPUTED         = False
options.DUMP_NEW_TASKS        = False
options.DUMP_YIELD_RESULTS    = False
options.DUMP_QUEUED_RESULTS   = False
options.DUMP_CONTEXTS         = False
options.DUMP_SYNC             = False
options.DUMP_STACK            = False  # When it's meaningful, e.g. on batch flush
options.DUMP_SCHEDULER_STATE  = False
options.DUMP_SYNC_CALLS       = False
options.COLLECT_PERF_STATS    = False

options.SCHEDULER_STATE_DUMP_INTERVAL = 1       # In seconds
options.DEBUG_STR_REPR_MAX_LENGTH     = 240     # In characters, 0 means infinity
options.STACK_DUMP_LIMIT              = 10      # In frames, None means infinity
options.MAX_TASK_STACK_SIZE           = 1000000 # Max number of concurrent futures + batch items

options.ENABLE_COMPLEX_ASSERTIONS = True

def DUMP_ALL(value=None):
    return options.DUMP_ALL(value)

# DUMP_ALL(True)


original_hook = None
is_attached = False
_std_str = str
_std_repr = repr


def dump_error(error, tb=None):
    """Dumps errors w/async stack traces."""
    try:
        stderr.write('\n' + (format_error(error, tb=tb) or 'No error'))
    finally:
        stdout.flush()
        stderr.flush()


def format_error(error, tb=None):
    """Formats errors w/async stack traces."""
    if error is None:
        return None
    result = ''
    if hasattr(error, '_traceback') or tb is not None:
        tb = tb or error._traceback
        result += ''.join(traceback.format_exception(error.__class__, error, tb))
    elif isinstance(error, BaseException):
        exc_text = ''.join(traceback.format_exception_only(error.__class__, error))
        if isinstance(exc_text, bytes):
            exc_text = exc_text.decode('utf-8', 'replace')
        result += '\n' + exc_text
    return result


class AsynqStackTracebackFormatter(logging.Formatter):
    """Prints traceback skipping asynq frames during logger.exception/error usages."""
    def formatException(self, exc_info):
        ty, val, tb = exc_info
        return format_error(val, tb=tb)


def _should_skip_frame(frame):
    if frame:
        traceback_hide_directive_name = '__traceback_hide__'
        return frame.f_locals.get(
            traceback_hide_directive_name,
            frame.f_globals.get(
                traceback_hide_directive_name,
                False)) is True
    return False


def extract_tb(tb, limit=None):
    """This implementation is stolen from traceback module but respects __traceback_hide__."""
    if limit is None:
        if hasattr(sys, 'tracebacklimit'):
            limit = sys.tracebacklimit
    tb_list = []
    n = 0
    while tb is not None and (limit is None or n < limit):
        f = tb.tb_frame
        if not _should_skip_frame(f):
            lineno = tb.tb_lineno
            co = f.f_code
            filename = co.co_filename
            name = co.co_name
            linecache.checkcache(filename)
            line = linecache.getline(filename, lineno, f.f_globals)
            if line:
                line = line.strip()
            else:
                line = None
            tb_list.append((filename, lineno, name, line))
        tb = tb.tb_next
        n = n + 1
    return tb_list


def format_tb(tb):
    """Formats a traceback into a list of lines."""
    return traceback.format_list(extract_tb(tb))


def dump_stack(skip=0, limit=None):
    """Dumps current stack trace."""
    skip += 2  # To skip dump_stack and traceback.extract_stack
    if limit is None:
        limit = options.STACK_DUMP_LIMIT
    print('--- Stack trace: -----------------------------------------------------')
    try:
        stack = traceback.extract_stack(limit=None if limit is None else limit + skip)
        print(''.join(traceback.format_list(stack[:-skip])), end='')
    finally:
        print('----------------------------------------------------------------------')
        stdout.flush()


def dump_asynq_stack():
    """Dumps the current asynq stack to stdout."""
    format_list = format_asynq_stack()
    if format_list is None:
        print('dump_asynq_stack: no asynq task currently active')
    else:
        print('\n'.join(format_list))


def format_asynq_stack():
    """Returns a list of strings.

    Each string corresponds to one item in the current asynq stack.

    Returns None if there is no active asynq task.

    """
    # Doing this in the global scope creates a circular dependency
    from .scheduler import get_scheduler
    active_task = get_scheduler().active_task
    if active_task is not None:
        return active_task.traceback()
    else:
        return None


def dump(state):
    if not options.DUMP_PRE_ERROR_STATE:
        return
    stdout.flush()
    stderr.flush()
    stdout.write('\n--- Pre-error state dump: --------------------------------------------\n')
    try:
        state.dump()
    finally:
        stdout.write('----------------------------------------------------------------------\n')
        stderr.write('\n')
        stdout.flush()
        stderr.flush()


def write(text, indent=0):
    if indent > 0:
        indent_str = '  ' * indent
        text = text.replace('\n', '\n' + indent_str)
        if not text.startswith('\n'):
            text = indent_str + text
    stdout.write(text + '\n')


def str(source, truncate=True):
    return qcore.safe_str(source, options.DEBUG_STR_REPR_MAX_LENGTH if truncate else 0)


def repr(source, truncate=True):
    return qcore.safe_repr(source, options.DEBUG_STR_REPR_MAX_LENGTH if truncate else 0)


def async_exception_hook(type, error, tb):
    """Exception hook capable of printing async stack traces."""
    global original_hook

    stdout.flush()
    stderr.flush()
    if original_hook is not None:
        original_hook(type, error, tb)
    dump_error(error, tb=tb)


def ipython_custom_exception_handler(self, etype, value, tb, tb_offset=None):
    """Override ipython's exception handler to print async traceback."""
    async_exception_hook(etype, value, tb)
    # below is the default exception handling behavior of ipython
    self.showtraceback()


def attach_exception_hook():
    """Injects async exception hook into the sys.excepthook."""
    try:
        # detect whether we're running in IPython
        __IPYTHON__
    except NameError:
        shell = None
    else:
        # override ipython's exception handler if in a shell.
        # we need to do this because ipython overrides sys.excepthook
        # so just the else block doesn't cover that case.
        from IPython.core.getipython import get_ipython

        # this may be None if __IPYTHON__ is somehow defined, but we are not
        # in fact in a shell
        shell = get_ipython()

    if shell is not None:
        shell.set_custom_exc((BaseException,), ipython_custom_exception_handler)
    else:
        global is_attached, original_hook
        if is_attached:
            sys.stderr.write("Warning: async exception hook was already attached.\n")
            return
        original_hook = sys.excepthook
        sys.excepthook = async_exception_hook
        is_attached = True


def detach_exception_hook():
    """Removes async exception hook into the sys.excepthook."""
    global is_attached, original_hook
    assert is_attached, "Async exception hook wasn't attached."
    sys.excepthook = original_hook
    is_attached = False


@contextlib.contextmanager
def enable_complex_assertions(enable=True):
    old = options.ENABLE_COMPLEX_ASSERTIONS
    try:
        options.ENABLE_COMPLEX_ASSERTIONS = enable
        yield None
    finally:
        options.ENABLE_COMPLEX_ASSERTIONS = old


def disable_complex_assertions():
    return enable_complex_assertions(False)


def sync():
    assert False, "'import asynq' seems broken: this function must be replaced with async.batching.sync."


def get_frame(generator):
    """Given a generator, returns its current frame."""
    if getattr(generator, 'gi_frame', None) is not None:
        return generator.gi_frame
    return None



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import asynq
from asynq.futures import ConstFuture

from qcore.asserts import assert_eq, assert_is, AssertRaises


def check_unwrap(expected, to_unwrap):
    # can't call unwrap() directly because it's hidden by cython
    @asynq()
    def caller():
        value = yield to_unwrap
        assert_eq(expected, value, extra='yielded {}'.format(to_unwrap))
    caller()


def test_unwrap():
    check_unwrap(None, None)
    check_unwrap(1, ConstFuture(1))

    for typ in (list, tuple):
        check_unwrap(typ(), typ())
        check_unwrap(typ([1]), typ([ConstFuture(1)]))
        check_unwrap(typ([1, 2]), typ([ConstFuture(1), ConstFuture(2)]))
        check_unwrap(typ([1, 2, 3]), typ([ConstFuture(1), ConstFuture(2), ConstFuture(3)]))

    with AssertRaises(TypeError):
        check_unwrap(1)



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
from cpython.ref cimport PyObject

cimport qcore.helpers as core_helpers
cimport qcore.inspection as core_inspection
cimport qcore.decorators
cimport futures
cimport async_task
cimport _debug

cdef _debug.DebugOptions _debug_options


# cpdef lazy(object fn)
cpdef bint has_async_fn(object fn) except -1
cpdef bint is_pure_async_fn(object fn) except -1


cdef class PureAsyncDecoratorBinder(qcore.decorators.DecoratorBinder):
    cpdef bint is_pure_async_fn(self) except -1


cdef class PureAsyncDecorator(qcore.decorators.DecoratorBase):
    cdef public type task_cls
    cdef public dict kwargs
    cdef public bint needs_wrapper

    cpdef str name(self)
    cpdef bint is_pure_async_fn(self) except -1
    cpdef object _call_pure(self, tuple args, dict kwargs)


cdef class AsyncDecoratorBinder(qcore.decorators.DecoratorBinder):
    pass


cdef class AsyncDecorator(PureAsyncDecorator):
    pass


cdef class AsyncAndSyncPairDecorator(AsyncDecorator):
    cdef public object sync_fn


cdef class AsyncProxyDecorator(AsyncDecorator):
    pass


cdef class AsyncAndSyncPairProxyDecorator(AsyncProxyDecorator):
    pass




# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qcore.asserts import AssertRaises
from six.moves import xrange

from asynq import asynq, debug, result
from .helpers import Profiler


def test_recursion():
    def sum(i, dump_progress=False):
        if i == 0:
            return 0
        if dump_progress and i % 100 == 0:
            print('... in sum(%s)' % i)
        return i + sum(i - 1, dump_progress)

    @asynq()
    def sum_async(i, dump_progress=False):
        if i == 0:
            result(0); return
        if dump_progress and i % 100 == 0:
            print('... in sum_async(%s)' % i)
        result(i + (yield sum_async.asynq(i - 1, dump_progress))); return

    with AssertRaises(RuntimeError):
        sum(2000, True)  # By default max stack depth is ~ 1000

    sum_async(2000, True)

    # But async is costly:

    p1 = Profiler("200 x sum(500)")
    with p1:
        for i in xrange(0, 200):
            sum(500)

    p2 = Profiler("20 x sum_async(500)() (w/assertions)")
    with p2:
        for i in xrange(0, 20):
            sum_async(500)

    print('Slowdown: %s' % (p2.diff * 10.0 / p1.diff))

    p3 = Profiler("20 x sum_async(500)() (w/o assertions)")
    with debug.disable_complex_assertions(), p3:
        for i in xrange(0, 20):
            sum_async(500)

    print('Slowdown: %s' % (p3.diff * 10.0 / p1.diff))

    p3 = Profiler("sum_async(10000)()")
    with p3:
        sum_async(10000)




# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sys import stderr, stdout
import time
import threading

from qcore import utime
import qcore.events as core_events


from . import debug
from . import futures
from . import batching
from . import _debug
from .async_task import AsyncTask

assert str(__name__).endswith('asynq.scheduler') \
    or str(__name__).endswith('asynq.lib64.scheduler'), \
    "Are you importing asynq from the wrong directory?"

_debug_options = _debug.options
_futures_none = futures._none


class AsyncTaskError(Exception):
    pass


class TaskScheduler(object):
    """Schedules and runs AsyncTask objects flushing batches when needed."""

    def __init__(self):
        self._last_dump_time = time.time()
        self.on_before_batch_flush = core_events.EventHook()
        self.on_after_batch_flush = core_events.EventHook()
        thread = threading.current_thread()
        thread_name = thread.name if thread.name else str(thread.ident)
        if '_state' in globals():
            _state.last_id += 1
            _id = _state.last_id
        else:
            _id = 0
        self.name = '%s / %d' % (thread_name, _id)
        self.reset()

    def reset(self):
        self._batches = set()
        self._tasks = []
        self.active_task = None

    def wait_for(self, task):
        """
        Executes a task and ensures it's complete when this method returns.

        :param tasks: task to wait for
        :return: ``None``
        """
        while not task.is_computed():
            self._execute(task)
            if task.is_computed():
                break
            self._continue_with_batch()

    def _execute(self, root_task):
        """Implements task execution loop.

        The postcondition of this method is that all tasks in the dependency tree of root_task
        that aren't blocked on batch items waiting to be flushed should be executed until they are
        (or until they're computed). This is done by running a depth-first search on the dependency
        tree.

        :param root_task: root of the dependency tree
        :return: ``None``
        """
        init_num_tasks = len(self._tasks)
        self._tasks.append(root_task)

        # Run the execution loop until the root_task is complete (it's either blocked on batch
        # items waiting to be flushed, or computed).
        while len(self._tasks) > init_num_tasks:
            if len(self._tasks) > _debug_options.MAX_TASK_STACK_SIZE:
                self.reset()
                debug.dump(self)
                raise RuntimeError('Number of scheduled tasks exceeded maximum threshold.')

            # _tasks is a stack, so take the last one.
            task = self._tasks[-1]
            if _debug_options.DUMP_SCHEDULER_STATE:
                self.try_time_based_dump()

            if task.is_computed():
                self._tasks.pop()
            elif isinstance(task, AsyncTask):
                self._handle_async_task(task)
            elif isinstance(task, batching.BatchItemBase):
                # This can happen multiple times per batch item (if we run _execute and this batch
                # item doesn't get flushed), but that's ok because self._batches is a set.
                self._schedule_batch(task.batch)
                self._tasks.pop()
            else:
                task._compute()
                self._tasks.pop()

    def _schedule_batch(self, batch):
        if batch.is_flushed():
            if _debug_options.DUMP_SCHEDULE_BATCH:
                debug.write("@async: can't schedule flushed batch %s" % debug.str(batch))
            return False
        if _debug_options.DUMP_SCHEDULE_BATCH and batch not in self._batches:
            debug.write('@async: scheduling batch %s' % debug.str(batch))
        self._batches.add(batch)
        return True

    def _flush_batch(self, batch):
        self.on_before_batch_flush(batch)
        try:
            batch.flush()
        finally:
            self.on_after_batch_flush(batch)
        return 0

    def _handle_async_task(self, task):
        # is_blocked indicates that one of the tasks dependencies isn't computed yet,
        # so we can't run _continue until they are.
        if task.is_blocked():
            # _dependencies_scheduled indicates if we've already added the task's
            # dependencies to the task stack. If the task is blocked and we've already
            # scheduled and run its dependencies, it's blocked on batch items waiting
            # to be flushed so we're done with this task.
            if task._dependencies_scheduled:
                # Set _dependencies_scheduled to false so on future runs of _execute,
                # we add the dependencies to the task stack (since some of the batch items
                # in the subtree might have been flushed)
                if _debug_options.DUMP_CONTINUE_TASK:
                    debug.write('@async: skipping %s' % debug.str(task))
                task._dependencies_scheduled = False
                task._pause_contexts()
                self._tasks.pop()
            # If the task is blocked and we haven't scheduled its dependencies, we
            # should do so now.
            else:
                task._dependencies_scheduled = True
                task._resume_contexts()
                for dependency in task._dependencies:
                    if not dependency.is_computed():
                        if _debug_options.DUMP_SCHEDULE_TASK:
                            debug.write('@async: scheduling task %s' % debug.str(dependency))
                        if _debug_options.DUMP_DEPENDENCIES:
                            debug.write('@async: +dependency: %s needs %s' % (debug.str(task), debug.str(dependency)))
                        self._tasks.append(dependency)
        else:
            self._continue_with_task(task)


    def _continue_with_task(self, task):
        task._resume_contexts()
        old_task = self.active_task
        self.active_task = task

        if _debug_options.DUMP_CONTINUE_TASK:
            debug.write('@async: -> continuing %s' % debug.str(task))
        if _debug_options.COLLECT_PERF_STATS:
            start = utime()
            task._continue()
            task._total_time += utime() - start
            if task.is_computed() and isinstance(task, AsyncTask):
                task.dump_perf_stats()
        else:
            task._continue()
        if _debug_options.DUMP_CONTINUE_TASK:
            debug.write('@async: <- continued %s' % debug.str(task))

        self.active_task = old_task
        # We get a new set of dependencies when we run _continue, so these haven't
        # been scheduled.
        task._dependencies_scheduled = False

    def _continue_with_batch(self):
        """
        Flushes one of batches (the longest one by default).

        :param assert_no_batch: indicates whether exception must be
                                raised if there is no batch to flush
        :return: the batch that was flushed, if there was a flush;
                 otherwise, ``None``.

        """
        batch = self._select_batch_to_flush()
        if batch is None:
            if _debug_options.DUMP_FLUSH_BATCH:
                debug.write('@async: no batch to flush')
            else:
                return None
        self._batches.remove(batch)
        self._flush_batch(batch)
        return batch

    def _select_batch_to_flush(self):
        """Returns the batch having highest priority,
        or ``None``, if there are no batches.

        This method uses ``BatchBase.get_priority()`` to
        determine the priority.

        Side effect: this method removed flushed batches.

        :return: selected batch or None.

        """
        best_batch = None
        best_priority = None
        batches_to_remove = None
        for batch in self._batches:
            if not batch.items or batch.is_flushed():
                if batches_to_remove is None:
                    batches_to_remove = [batch]
                else:
                    batches_to_remove.append(batch)
                continue
            priority = batch.get_priority()
            if best_batch is None or best_priority < priority:
                best_batch = batch
                best_priority = priority
        if batches_to_remove:
            for batch in batches_to_remove:
                self._batches.remove(batch)
        return best_batch

    def __str__(self):
        return '%s %s (%d tasks, %d batches; active task: %s)' % \
            (type(self), repr(self.name), len(self._tasks), len(self._batches), str(self.active_task))

    def __repr__(self):
        return self.__str__()

    def dump(self, indent=0):
        debug.write(debug.str(self), indent)
        if self._tasks:
            debug.write('Task queue:', indent + 1)
            for task in self._tasks:
                task.dump(indent + 2)
        else:
            debug.write('No tasks in task queue.', indent + 1)
        if self._batches:
            debug.write('Batches:', indent + 1)
            for batch in self._batches:
                batch.dump(indent + 2)

    def try_time_based_dump(self, last_task=None):
        current_time = time.time()
        if (current_time - self._last_dump_time) < _debug_options.SCHEDULER_STATE_DUMP_INTERVAL:
            return
        self._last_dump_time = current_time
        debug.write('\n--- Scheduler state dump: --------------------------------------------')
        try:
            self.dump()
            if last_task is not None:
                debug.write('Last task: %s' % debug.str(last_task), 1)
        finally:
            debug.write('----------------------------------------------------------------------\n')
            stdout.flush()
            stderr.flush()

class LocalTaskSchedulerState(threading.local):
    def __init__(self):
        self.last_id = 0
        self.reset()

    def reset(self):
        self.current = TaskScheduler()

_state = LocalTaskSchedulerState()
globals()['_state'] = _state


def get_scheduler():
    global _state
    return _state.current

def reset():
    _state.reset()

def get_active_task():
    global _state
    s = _state.current
    return None if s is None else s.active_task



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
from cpython.ref cimport PyObject

cimport qcore.inspection as core_inspection
cimport futures
cimport _debug


cdef _debug.DebugOptions _debug_options


cdef class BatchBase(futures.FutureBase):
    cdef public list items

    cpdef bint is_flushed(self) except -1
    cpdef bint is_cancelled(self) except -1
    cpdef bint is_empty(self) except -1
    cpdef object get_priority(self)

    cpdef flush(self)
    cpdef cancel(self, object error=?)

    cpdef _compute(self)
    cpdef _computed(self)
    cpdef _flush(self)
    cpdef _cancel(self)
    cpdef _try_switch_active_batch(self)

    cpdef dump(self, int indent=?)

cdef class BatchItemBase(futures.FutureBase):
    cdef public BatchBase batch
    cdef public long index

    cpdef _compute(self)


cdef class DebugBatchItem(BatchItemBase):
    cdef public object _result

cdef class DebugBatch(BatchBase):
    cdef public str name
    cdef public long index

    cpdef _flush(self)
    cpdef _cancel(self)


cdef object _debug_batch_state

cpdef sync(tag=?)



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import asynq, result
from .debug_cache import reset_caches, mc
from .caching import ExternalCacheBatchItem

def test_chain():
    @asynq()
    def foo(num_yield):
        if num_yield == 0:
            result(0); return

        yield ExternalCacheBatchItem(mc._batch, 'get', 'test')
        yield foo.asynq(num_yield - 1)

    reset_caches()
    foo(10)
    assert mc._batch.index == 11


def test_tree():
    @asynq()
    def foo(depth):
        if depth == 0:
            result((yield ExternalCacheBatchItem(mc._batch, 'get', 'test'))); return
        yield foo.asynq(depth - 1), foo.asynq(depth - 1)

    reset_caches()
    foo(5)
    assert mc._batch.index == 2



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
from cpython.ref cimport PyObject

cimport qcore.events as core_events
cimport futures
cimport async_task
cimport batching
cimport _debug


cdef _debug.DebugOptions _debug_options
cdef object _futures_none

cdef class TaskScheduler(object):
    cdef public float _last_dump_time
    cdef public set _batches
    cdef public core_events.EventHook on_before_batch_flush
    cdef public core_events.EventHook on_after_batch_flush
    cdef public async_task.AsyncTask active_task
    cdef public str name
    cdef public list _tasks

    cdef reset(self)

    cpdef int wait_for(self, async_task.AsyncTask task) except -1
    cdef int _execute(self, async_task.AsyncTask root_task) except -1

    cdef _schedule_batch(self, batching.BatchBase batch)
    cdef int _flush_batch(self, batching.BatchBase batch) except -1

    cdef _handle_async_task(self, async_task.AsyncTask task)
    cdef int _continue_with_task(self, async_task.AsyncTask task) except -1
    cdef batching.BatchBase _continue_with_batch(self)
    cdef batching.BatchBase _select_batch_to_flush(self)

    cpdef dump(self, int indent=?)
    @cython.locals(current_time=float)
    cpdef try_time_based_dump(self, object last_task=?)

cdef object _state

cpdef TaskScheduler get_scheduler()

cpdef async_task.AsyncTask get_active_task()



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qcore import MarkerObject
from asynq import asynq, async_proxy, result, ConstFuture
from collections import deque


# TODO(alex): finish w/this test


empty = MarkerObject(u'empty @ asynq.channels')
future_empty = ConstFuture(empty)
future_false = ConstFuture(False)
future_true = ConstFuture(True)


class Channel(object):
    def __init__(self, capacity=256):
        self.capacity = capacity
        self.items = deque()
        self.futures = deque()

    @async_proxy(pure=True)
    def push(self, value, should_await=True):
        if self.futures:
            future = self.futures.popleft()
            future.set_value(value)
            return future_true
        if len(self.items) < self.capacity:
            self.items.append(value)
            return future_true
        return _push_async(self, value) \
            if should_await else future_false

    @async_proxy(pure=True)
    def pull(self, should_await=True):
        if self.items:
            return ConstFuture(self.items.popleft())
        return _pull_async(self) \
            if should_await else future_empty


@asynq(pure=True)
def _push_async(channel, value):
    yield
    while True:
        if channel.push(value, False) is future_true:
            result(future_true); return
        yield


@asynq(pure=True)
def _pull_async(channel):
    yield
    while True:
        item = channel.pull(False)
        if item is not future_empty:
            result(item); return
        yield



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup
from setuptools.extension import Extension

import codecs
import sys


CYTHON_MODULES = [
    'async_task',
    'batching',
    'contexts',
    '_debug',
    'decorators',
    'futures',
    'profiler',
    'scheduler',
    'scoped_value',
    'utils',
]


DATA_FILES = ['%s.pxd' % module for module in CYTHON_MODULES]


VERSION = '1.0.1'


EXTENSIONS = [
    Extension(
        'asynq.%s' % module,
        ['asynq/%s.py' % module]
    ) for module in CYTHON_MODULES
]


if __name__ == '__main__':
    with codecs.open('./README.rst', encoding='utf-8') as f:
        long_description = f.read()

    requirements = ['Cython', 'qcore', 'setuptools', 'inspect2']
    if sys.version_info < (3, 3):
        # mock is in the standard library since Python 3.3
        requirements.append('mock')

    setup(
        name='asynq',
        version=VERSION,
        author='Quora, Inc.',
        author_email='asynq@quora.com',
        description='Quora\'s asynq library',
        long_description=long_description,
        url='https://github.com/quora/asynq',
        license='Apache Software License',
        classifiers=[
            'License :: OSI Approved :: Apache Software License',

            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
        ],
        keywords='quora asynq common utility',
        packages=['asynq', 'asynq.tests'],
        package_data={'asynq': DATA_FILES},
        ext_modules=EXTENSIONS,
        setup_requires=['Cython'],
        install_requires=[
            'Cython',
            'qcore',
            'setuptools',
            'inspect2',
            'mock; python_version < "3.3"',
        ],
    )



import asynq
import pickle
from qcore.asserts import assert_eq, AssertRaises


def test_constfuture_pickling():
    fut = asynq.ConstFuture(3)

    for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
        pickled = pickle.dumps(fut, protocol)
        assert_eq(fut.value(), pickle.loads(pickled).value())


def test_callback_exception_handling():
    def raise_exception(task):
        raise ValueError

    def base_exception(task):
        raise KeyboardInterrupt

    fut_exc = asynq.Future(lambda: 3)
    fut_exc.on_computed.subscribe(raise_exception)

    fut_base_exc = asynq.Future(lambda: 4)
    fut_base_exc.on_computed.subscribe(base_exception)

    # exception is passed on and value gets set correctly
    assert_eq(3, fut_exc.value())

    # since there's a base exception, we fail when we try to set the value to the future
    with AssertRaises(KeyboardInterrupt):
        assert_eq(4, fut_base_exc.value())



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .futures import FutureBase
from . import async_task
from . import _debug


_debug_options = _debug.options

def result(value):
    """An async equivalent for return for async methods: result(x); return."""
    assert not isinstance(value, FutureBase), \
        "You probably forgot to yield this value before returning"
    raise async_task.AsyncTaskResult(value)




# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qcore.asserts import assert_eq, assert_is, assert_is_instance, AssertRaises
from asynq import asynq, async_proxy, is_pure_async_fn, result, async_call, ConstFuture
from asynq.decorators import lazy, get_async_fn, get_async_or_sync_fn, make_async_decorator, AsyncDecorator
import pickle


def double_return_value(fun):
    @asynq(pure=True)
    def wrapper_fn(*args, **kwargs):
        value = yield fun.asynq(*args, **kwargs)
        result(value * 2); return
    return make_async_decorator(fun, wrapper_fn, 'double_return_value')


@double_return_value
@asynq()
def square(x):
    return x * x


class MyClass(object):
    @asynq()
    def method(self, number):
        assert type(self) is MyClass
        cls, one = self.get_cls_and_args(number)
        assert cls is MyClass
        assert one == number
        cls, one = yield self.get_cls_and_args.asynq(number)
        assert cls is MyClass
        assert one == number
        one = yield self.static(number)
        assert one == number
        one = yield self.static_ac.asynq(number)
        assert one == number
        cls, proxied = yield self.async_proxy_classmethod.asynq(number)
        assert cls is MyClass
        assert proxied == number
        result(self); return

    @async_proxy()
    @classmethod
    def async_proxy_classmethod(cls, number):
        return cls.get_cls_and_args.asynq(number)

    @asynq()
    @classmethod
    def get_cls_and_args(cls, number):
        print('get_cls_and_args')
        assert cls.get_cls_ac() is cls
        assert (yield cls.get_cls_ac.asynq()) is cls
        assert cls.get_cls().value() is cls
        assert (yield cls.get_cls()) is cls
        result((cls, number)); return

    @asynq()
    @classmethod
    def get_cls_ac(cls):
        print('get_cls_ac')
        result(cls); return

    @asynq(pure=True)
    @classmethod
    def get_cls(cls):
        print('get_cls')
        result(cls); return

    @asynq()
    @staticmethod
    def static_ac(number):
        print('static_ac')
        result(number); return

    @staticmethod
    @asynq(pure=True)
    def static(number):
        print('static')
        result(number); return

    @staticmethod
    def sync_staticmethod():
        return 'sync_staticmethod'

    @asynq(sync_fn=sync_staticmethod)
    @staticmethod
    def async_staticmethod():
        return 'async_staticmethod'

    @classmethod
    def sync_classmethod(cls):
        return 'sync_classmethod'

    @asynq(sync_fn=sync_classmethod)
    @classmethod
    def async_classmethod(cls):
        return 'async_classmethod'

    def sync_method(self):
        return 'sync_method'

    @asynq(sync_fn=sync_method)
    def async_method(self):
        return 'async_method'

    @double_return_value
    @asynq()
    @classmethod
    def square(cls, x):
        return x * x


def sync_fn():
    return 'sync_fn'


@asynq(sync_fn=sync_fn)
def async_fn():
    return 'async_fn'


@asynq(pure=True)
def pure_async_fn():
    return 'pure_async_fn'


def sync_proxied_fn():
    return 'sync_proxied_fn'


@async_proxy(sync_fn=sync_proxied_fn)
def async_proxied_fn():
    return ConstFuture('async_proxied_fn')


@lazy
def lazy_fn(a, b):
    return a + b


def test_lazy():
    future = lazy_fn(1, 2)
    assert not future.is_computed()
    assert_eq(3, future.value())
    assert future.is_computed()


class DisallowSetting(object):
    def fn(self):
        return False

    def __setattr__(self, attr, value):
        raise AttributeError('cannot set attribute %s' % attr)


def test_is_pure_async_fn():
    assert is_pure_async_fn(lazy_fn)
    assert not is_pure_async_fn(test_lazy)
    assert not is_pure_async_fn(async_fn)
    assert is_pure_async_fn(pure_async_fn)
    assert not is_pure_async_fn(DisallowSetting())
    assert is_pure_async_fn(MyClass.get_cls)
    assert not is_pure_async_fn(MyClass.get_cls_ac)
    assert not is_pure_async_fn(AsyncDecorator)


def test_get_async_fn():
    assert_eq(async_fn.asynq, get_async_fn(async_fn))
    assert_eq(lazy_fn, get_async_fn(lazy_fn))
    assert_is(None, get_async_fn(sync_fn))

    wrapper = get_async_fn(sync_fn, wrap_if_none=True)
    assert is_pure_async_fn(wrapper)
    result = wrapper()
    assert_is_instance(result, ConstFuture)
    assert_eq('sync_fn', result.value())


def test_get_async_or_sync_fn():
    assert_is(sync_fn, get_async_or_sync_fn(sync_fn))
    assert_eq(async_fn.asynq, get_async_or_sync_fn(async_fn))


def test_async_proxy():
    assert_eq('sync_proxied_fn', sync_proxied_fn())
    assert_eq('sync_proxied_fn', async_proxied_fn())

    result = async_proxied_fn.asynq()
    assert_is_instance(result, ConstFuture)
    assert_eq('async_proxied_fn', result.value())

    with AssertRaises(AssertionError):
        @async_proxy(pure=True, sync_fn=sync_proxied_fn)
        def this_doesnt_make_sense():
            pass


def test():
    obj = MyClass()
    assert obj is obj.method(1)


def test_staticmethod_sync_fn():
    assert_eq('sync_staticmethod', MyClass.async_staticmethod())
    assert_eq('async_staticmethod', MyClass.async_staticmethod.asynq().value())


def test_classmethod_sync_fn():
    assert_eq('async_classmethod', MyClass.async_classmethod.asynq().value())
    assert_eq('sync_classmethod', MyClass.async_classmethod())


def test_method_sync_fn():
    instance = MyClass()
    assert_eq('sync_method', instance.async_method())
    assert_eq('async_method', instance.async_method.asynq().value())


def test_pickling():
    pickled = pickle.dumps(async_fn)
    unpickled = pickle.loads(pickled)
    assert_eq('sync_fn', unpickled())


def test_async_call():
    @asynq()
    def f1(arg, kw=1):
        return arg, kw

    @asynq(pure=True)
    def f2(arg, kw=1):
        return arg, kw

    def f3(arg, kw=1):
        return arg, kw

    for f in [f1, f2, f3]:
        assert_eq((10, 1), async_call.asynq(f, 10).value())
        assert_eq((10, 5), async_call.asynq(f, 10, 5).value())
        assert_eq((10, 7), async_call.asynq(f, 10, kw=7).value())


def test_make_async_decorator():
    assert_eq(18, square(3))
    assert_eq(18, MyClass.square(3))
    assert_eq(18, square.asynq(3).value())
    assert_eq(18, MyClass.square.asynq(3).value())

    assert not is_pure_async_fn(square)
    assert_eq('@double_return_value()', square.name())



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qcore.asserts import assert_eq
from asynq import AsyncScopedValue, asynq, result, async_override
from asynq.batching import DebugBatchItem

v = AsyncScopedValue('a')


@asynq()
def async_scoped_value_helper(inner_val):
    @asynq()
    def nested():
        assert_eq(v.get(), inner_val)
        yield DebugBatchItem()
        with v.override('c'):
            yield DebugBatchItem()  # just so other function gets scheduled
            assert_eq(v.get(), 'c')
            yield DebugBatchItem()

    assert_eq(v.get(), 'a')
    yield DebugBatchItem()
    with v.override(inner_val):
        yield DebugBatchItem()
        assert_eq(v.get(), inner_val)
        result((yield nested.asynq())); return


@asynq()
def async_scoped_value_caller():
    yield async_scoped_value_helper.asynq('e'), async_scoped_value_helper.asynq('f')


def test_async_scoped_value():

    async_scoped_value_caller()

    val = AsyncScopedValue('capybara')
    assert_eq('capybara', val.get())
    val.set('nutria')
    assert_eq('nutria', val.get())

    assert_eq('AsyncScopedValue(nutria)', str(val))
    assert_eq("AsyncScopedValue('nutria')", repr(val))


def test_exception():
    @asynq()
    def test_body():
        assert_eq(v(), 'a')
        yield
        try:
            with v.override('b'):
                yield
                assert_eq(v(), 'b')
                yield
                raise NotImplementedError()
        except NotImplementedError:
            yield
            pass
        yield
        assert_eq(v(), 'a')

    test_body()


def test_override():
    class TestObject(object):
        def __init__(self):
            self.v = None

    o = TestObject()
    o.v = 'a'

    @asynq()
    def test_body():
        assert_eq(o.v, 'a')
        yield
        with async_override(o, 'v', 'b'):
            assert_eq(o.v, 'b')
            yield
            try:
                with async_override(o, 'v', 'c'):
                    assert_eq(o.v, 'c')
                    yield
                    raise NotImplementedError()
            except NotImplementedError:
                pass
            assert_eq(o.v, 'b')
        yield
        assert_eq(o.v, 'a')

    test_body()



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import qcore.helpers as core_helpers
import qcore.inspection as core_inspection
import qcore.decorators
import sys

from . import futures
from . import async_task
from . import _debug

__traceback_hide__ = True

_debug_options = _debug.options


def lazy(fn):
    """Converts a function into a lazy one - i.e. its call
    returns a Future bound to a function call with passed
    arguments, that can be evaluated on demand later

    """
    @core_inspection.wraps(fn)
    def new_fn(*args, **kwargs):
        value_provider = lambda: fn(*args, **kwargs)
        return futures.Future(value_provider)

    new_fn.is_pure_async_fn = core_helpers.true_fn
    return new_fn


def has_async_fn(fn):
    """Returns true if fn can be called asynchronously."""
    return hasattr(fn, 'async') or hasattr(fn, 'asynq')


def is_pure_async_fn(fn):
    """Returns true if fn is an @asynq(pure=True) or @async_proxy(pure=True) function."""
    if hasattr(fn, 'is_pure_async_fn'):
        try:
            return fn.is_pure_async_fn()
        except (IndexError, TypeError):
            # This happens when calling e.g. is_pure_async_fn(AsyncDecorator). TypeError is the
            # expected exception in this case, but with some versions of Cython IndexError is thrown
            # instead.
            return False
    if hasattr(fn, 'fn'):
        result = is_pure_async_fn(fn.fn)
        result_fn = core_helpers.true_fn if result else core_helpers.false_fn
        try:
            fn.is_pure_async_fn = result_fn
        except (TypeError, AttributeError):
            pass  # some callables don't let you assign attributes
        return result
    return False


def is_async_fn(fn):
    """Returns true if fn is an @asynq([pure=True]) or @async_proxy(pure=True) function."""
    return hasattr(fn, 'asynq') or hasattr(fn, 'async') or is_pure_async_fn(fn)


def get_async_fn(fn, wrap_if_none=False):
    """Returns an async function for the specified source function."""
    if hasattr(fn, 'asynq'):
        return fn.asynq
    if hasattr(fn, 'async'):
        return getattr(fn, 'async')
    if is_pure_async_fn(fn):
        return fn
    if wrap_if_none:
        def sync_to_async_fn_wrapper(*args, **kwargs):
            return futures.ConstFuture(fn(*args, **kwargs))

        sync_to_async_fn_wrapper.is_pure_async_fn = core_helpers.true_fn
        return sync_to_async_fn_wrapper
    return None


def get_async_or_sync_fn(fn):
    """Returns an async function for the specified fn, if it exists; otherwise returns source."""
    if hasattr(fn, 'asynq'):
        return fn.asynq
    if hasattr(fn, 'async'):
        return getattr(fn, 'async')
    return fn


class PureAsyncDecoratorBinder(qcore.decorators.DecoratorBinder):
    def is_pure_async_fn(self):
        return True


class PureAsyncDecorator(qcore.decorators.DecoratorBase):
    binder_cls = PureAsyncDecoratorBinder

    def __init__(self, fn, task_cls, kwargs={}):
        qcore.decorators.DecoratorBase.__init__(self, fn)
        self.task_cls = task_cls
        self.needs_wrapper = core_inspection.is_cython_or_generator(fn)
        self.kwargs = kwargs

    def name(self):
        return '@asynq(pure=True)'

    def is_pure_async_fn(self):
        return True

    def _fn_wrapper(self, args, kwargs):
        raise async_task.AsyncTaskResult(self.fn(*args, **kwargs)); return
        yield

    def __call__(self, *args, **kwargs):
        return self._call_pure(args, kwargs)

    def _call_pure(self, args, kwargs):
        if not self.needs_wrapper:
            result = self._fn_wrapper(args, kwargs)
        else:
            result = self.fn(*args, **kwargs)
        return self.task_cls(result, self.fn, args, kwargs, **self.kwargs)


class AsyncDecoratorBinder(qcore.decorators.DecoratorBinder):
    def asynq(self, *args, **kwargs):
        if self.instance is None:
            return self.decorator.asynq(*args, **kwargs)
        else:
            return self.decorator.asynq(self.instance, *args, **kwargs)

    async = asynq


class AsyncDecorator(PureAsyncDecorator):
    binder_cls = AsyncDecoratorBinder

    def is_pure_async_fn(self):
        return False

    def asynq(self, *args, **kwargs):
        return self._call_pure(args, kwargs)

    async = asynq

    def name(self):
        return '@asynq()'

    def __call__(self, *args, **kwargs):
        return self._call_pure(args, kwargs).value()


class AsyncAndSyncPairDecoratorBinder(AsyncDecoratorBinder):
    def __call__(self, *args, **kwargs):
        # the base class implementation adds .instance here, but we don't want that because we
        # called __get__ on the sync_fn manually; if we don't do this we'll end up adding self or
        # cls twice
        return self.decorator(*args, **kwargs)


class AsyncAndSyncPairDecorator(AsyncDecorator):
    binder_cls = AsyncAndSyncPairDecoratorBinder

    def __init__(self, fn, cls, sync_fn, kwargs={}):
        AsyncDecorator.__init__(self, fn, cls, kwargs)
        self.sync_fn = sync_fn

    def __call__(self, *args, **kwargs):
        return self.sync_fn(*args, **kwargs)

    def __get__(self, owner, cls):
        # This is needed so that we can use objects with __get__ as the sync_fn. If we just rely on
        # the base class's __get__ implementation, we'll never end up calling __get__ on the
        # sync_fn, so if it's a method of some sort it will not be bound correctly. We get around
        # this by manually calling __get__ on the sync_fn, then creating a copy of ourselves with
        # the bound sync_fn and calling the base class's __get__ implementation on it.
        sync_fn = self.sync_fn.__get__(owner, cls)
        fn = self.fn
        if self.type in (staticmethod, classmethod):
            fn = self.type(fn)
        new_self = qcore.decorators.decorate(
            AsyncAndSyncPairDecorator, self.task_cls, sync_fn, self.kwargs,
        )(fn)
        return AsyncDecorator.__get__(new_self, owner, cls)


class AsyncProxyDecorator(AsyncDecorator):
    def __init__(self, fn):
        # we don't need the task class but still need to pass it to the superclass
        AsyncDecorator.__init__(self, fn, None)

    def _call_pure(self, args, kwargs):
        return self.fn(*args, **kwargs)


class AsyncAndSyncPairProxyDecorator(AsyncProxyDecorator):
    def __init__(self, fn, sync_fn):
        AsyncProxyDecorator.__init__(self, fn)
        self.sync_fn = sync_fn

    def __call__(self, *args, **kwargs):
        return self.sync_fn(*args, **kwargs)


def asynq(pure=False, sync_fn=None, cls=async_task.AsyncTask, **kwargs):
    """Async task decorator.
    Converts a method returning generator object to
    a method returning AsyncTask object.

    """
    if kwargs:
        assert pure, "custom kwargs are only supported with pure=True"
    if pure:
        assert sync_fn is None, "sync_fn is not supported for pure async functions"

    def decorate(fn):
        assert not (is_pure_async_fn(fn) or has_async_fn(fn)), \
            "@asynq() decorator can be applied just once"
        if pure:
            return qcore.decorators.decorate(PureAsyncDecorator, cls, kwargs)(fn)
        elif sync_fn is None:
            return qcore.decorators.decorate(AsyncDecorator, cls)(fn)
        else:
            return qcore.decorators.decorate(AsyncAndSyncPairDecorator, cls, sync_fn)(fn)

    return decorate


if sys.version_info <= (3, 7):
    globals()['async'] = asynq


def async_proxy(pure=False, sync_fn=None):
    if sync_fn is not None:
        assert pure is False, "sync_fn=? cannot be used together with pure=True"

    def decorate(fn):
        if pure:
            return fn
        if sync_fn is None:
            return qcore.decorators.decorate(AsyncProxyDecorator)(fn)
        else:
            return qcore.decorators.decorate(AsyncAndSyncPairProxyDecorator, sync_fn)(fn)

    return decorate


@async_proxy()
def async_call(fn, *args, **kwargs):
    """Use this if you are not sure if fn is async or not.

    e.g. when you are within an async function and you need to call fn but it could either be
    async or non-async, you should write
    val = yield async_call.asynq(fn, arg1, kw1=value1)

    """
    if is_pure_async_fn(fn):
        return fn(*args, **kwargs)
    elif hasattr(fn, 'asynq'):
        return fn.asynq(*args, **kwargs)
    elif hasattr(fn, 'async'):
        return getattr(fn, 'async')(*args, **kwargs)
    else:
        return futures.ConstFuture(fn(*args, **kwargs))


class AsyncWrapper(qcore.decorators.DecoratorBase):
    """Implements make_async_decorator."""
    binder_cls = AsyncDecoratorBinder

    def __init__(self, fn, wrapper_fn, name):
        super(AsyncWrapper, self).__init__(fn)
        self.wrapper_fn = wrapper_fn
        self.function_name = name

    def name(self):
        return '@%s()' % self.function_name

    def _call_async(self, args, kwargs):
        return self.wrapper_fn(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self._call_async(args, kwargs).value()

    def asynq(self, *args, **kwargs):
        return self._call_async(args, kwargs)

    async = asynq

    def is_pure_async_fn(self):
        return False


def make_async_decorator(fn, wrapper_fn, name):
    """For implementing decorators that wrap async functions.

    Arguments:
    - fn: the function to be wrapped
    - wrapper_fn: function used to implement the decorator. Must return a Future.
    - name: name of the decorator (used for introspection)

    """
    return qcore.decorators.decorate(
        AsyncWrapper,
        wrapper_fn,
        name,
    )(fn)




# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import threading
import qcore.inspection as core_inspection

from . import debug
from . import futures
from . import _debug


__traceback_hide__ = True

_debug_options = _debug.options


class BatchingError(Exception):
    pass


class BatchCancelledError(BatchingError):
    pass


class BatchBase(futures.FutureBase):
    """Abstract base class describing a batch of operations."""
    def __init__(self):
        futures.FutureBase.__init__(self)  # Cython doesn't support super(...)
        self.items = []

    def is_flushed(self):
        return self.is_computed()

    def is_cancelled(self):
        return self.is_computed() and self.error() is not None

    def is_empty(self):
        return len(self.items) == 0

    def get_priority(self):
        """Returns batch flush priority.
        The higher priority, the earlier batch is flushed.

        By default it returns ``(0, len(self.items))`` tuple;
        0 here is base batch priority (zero by default).

        :return: batch flush priority.

        """
        return 0, len(self.items)

    def flush(self):
        """Flushes the batch.

        Almost the same as ``self.value()``, but:

        * this method doesn't throw an error even if underlying batch
          flush actually completed with an error
        * on the other hand, subsequent flush throws an error.

        So this method is intended to be called by schedulers:

        * They must flush each batch just once
        * They don't care (and moreover, shouldn't know) about actual
          flush errors. These errors will be anyway re-thrown later -
          on attempt to access values of underlying batch items.

        """
        if self.is_computed():
            raise BatchingError('Batch is already flushed or cancelled.')
        if _debug_options.DUMP_FLUSH_BATCH:
            debug.write('@async: -> batch flush:')
            self.dump(4)
            if _debug_options.DUMP_STACK:
                debug.dump_stack()
        try:
            self.error()  # Makes future to compute w/o raising an error
        finally:
            if _debug_options.DUMP_FLUSH_BATCH:
                debug.write('@async: <- batch flushed: %s' % debug.str(self))

    def cancel(self, error=None):
        """Cancels the batch.

        It's a public ``_cancel()`` wrapper enforcing
        additional safety properties.

        """
        if self.is_computed():
            return  # Cancel must never raise an error
        if error is None:
            error = BatchCancelledError()
        self.set_error(error)

    def _compute(self):
        self._try_switch_active_batch()
        try:
            self._flush()
            self.set_value(None)
        except BaseException as error:
            if not self.is_computed():
                self.set_error(error)
            raise  # Must re-throw to properly implement FutureBase API

    def _computed(self):
        # The purpose of this overridden method is to ensure that
        # all items are computed before on_computed event is raised.
        self._try_switch_active_batch()
        error = self.error()
        cancelled = error is not None
        if cancelled:
            self._cancel()
        for item in self.items:
            if not item.is_computed():
                # We must ensure all batch items are computed
                item.set_error(
                    error if cancelled
                    else AssertionError("Value of this item wasn't set on batch flush."))
        futures.FutureBase._computed(self)  # Cython doesn't support super(...)

    def _flush(self):
        """A protected method that must be override to implement batch flush.
        Normally it should simply forward the call to appropriate service
        (cache, DB, etc.), that must execute the batch (i.e. ensure each
        batch item will be able to acquire its value on subsequent attempt)
        and set its current batch to the newly created one.

        """
        raise NotImplementedError()

    def _cancel(self):
        """A protected method that must be override to implement batch cancellation.
        Normally it should simply forward the call to appropriate service
        (cache, DB, etc.), that must discard the current batch in this case.

        This method is optional to implement: _computed implementation
        anyway ensures that all items are computed as well, but
        you can add some additional logic here, if you want to.

        """
        pass

    def _try_switch_active_batch(self):
        """This protected method must be overridden to switch to a new active
        batch of this type.

        Must never throw an error.

        """
        raise NotImplementedError()

    def __str__(self):
        return '%s (%s, %i items)' % (
            core_inspection.get_full_name(type(self)),
            'cancelled' if self.is_cancelled() else
                'flushed' if self.is_flushed() else 'pending',
            len(self.items))

    def dump(self, indent=0):
        debug.write(debug.str(self), indent)
        debug.write('Priority: %s' % debug.repr(self.get_priority()), indent + 1)
        if self.items:
            debug.write('Items:', indent + 1)
            for item in self.items:
                item.dump(indent + 2)
        else:
            debug.write('No items.', indent + 1)


class BatchItemBase(futures.FutureBase):
    """Abstract base class describing batch item.
    Batch items are futures providing result
    of a particular cache operation.

    """
    def __init__(self, batch):
        super(BatchItemBase, self).__init__()
        assert not batch.is_flushed(), "can't add an item to the batch that is already flushed"
        self.batch = batch
        self.index = len(batch.items)
        batch.items.append(self)

    def _compute(self):
        """This method ensures the value is available
        by flushing the batch, if necessary.

        """
        if not self.batch.is_flushed():
            self.batch.flush()


class DebugBatchItem(BatchItemBase):
    """Debug batch item used to sync async execution."""

    def __init__(self, batch_name='default', result=None):
        global _debug_batch_state
        batch = _debug_batch_state.batches.setdefault(batch_name, DebugBatch(batch_name))
        super(DebugBatchItem, self).__init__(batch)
        self._result = result


class DebugBatch(BatchBase):
    """Debug batch used to sync async execution."""

    def __init__(self, name='default', index=0):
        super(DebugBatch, self).__init__()
        self.name = name
        self.index = index

    def _try_switch_active_batch(self):
        if _debug_batch_state.batches.get(self.name, None) is self:
            _debug_batch_state.batches[self.name] = DebugBatch(self.name, self.index + 1)

    def _flush(self):
        global _debug_batch_state
        if _debug_options.DUMP_SYNC:
            debug.write("@async.debug.sync: flushing batch %s (%i)" % (debug.repr(self.name), self.index))
        for item in self.items:
            item.set_value(item._result)

    def _cancel(self):
        global _debug_batch_state
        if _debug_options.DUMP_SYNC:
            debug.write("@async.debug.sync: cancelling batch %s (%i)" % (debug.repr(self.name), self.index))


class LocalDebugBatchState(threading.local):
    def __init__(self):
        super(LocalDebugBatchState, self).__init__()
        self.batches = {}

_debug_batch_state = LocalDebugBatchState()
globals()['_debug_batch_state'] = _debug_batch_state


def sync(tag='default'):
    return DebugBatchItem('sync-' + tag)

